using Test, BenchmarkTools
using LinearAlgebra
using DualTVDD:
    DualTVL1ROFOpProblem, ProjGradAlgorithm, ChambolleAlgorithm, DualTVDDAlgorithm,
    init, step!, fetch_u

@testset "B = I" begin
    g = Float64[0 2; 1 0]
    prob = DualTVL1ROFOpProblem(g, I, 1e-10)

    @testset for alg in (ProjGradAlgorithm(prob, τ=1/8), ChambolleAlgorithm(prob))
        ctx = init(alg)
        #@test 0 == @ballocated step!($ctx)
        for i in 1:100
            step!(ctx)
        end
        u = fetch_u(ctx)
        @test u ≈ g
    end
end

@testset "B = rand(...)" begin
    g = Float64[0 2; 1 0]
    B = rand(length(g), length(g))
    prob = DualTVL1ROFOpProblem(g, B, 1e-10)

    @testset for alg in (ProjGradAlgorithm(prob, τ=1/8), ChambolleAlgorithm(prob))
        ctx = init(alg)
        #@test 0 == @ballocated step!($ctx)
        for i in 1:100
            step!(ctx)
        end
        u = fetch_u(ctx)
        @test vec(u) ≈ B * vec(g)
    end
end

@testset "DualTVDDAlgorithm" begin
    n = 5
    ninner = 100
    g = rand(n, n)
    B = Diagonal(rand(n^2))
    # big λ is ok, since we test for inter-subdomain communication
    prob = DualTVL1ROFOpProblem(g, B, 100.)

    algref = ChambolleAlgorithm(prob)
    alg = DualTVDDAlgorithm(prob; M=(2,2), overlap=(2,2), ninner, parallel = false, σ = 0.25, subalg = x -> ChambolleAlgorithm(x))

    stref = init(algref)
    st = init(alg)
    #@test 0 == @ballocated step!($ctx)
    for i in 1:1000*ninner
        step!(stref)
    end
    for i in 1:1000
        step!(st)
    end
    @test fetch_u(st) ≈ fetch_u(stref)
end
